package ch.epfl.collections;

import java.util.Arrays;
import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Array-based implementation of the generic List interface. 
 *
 */
public class ArrayList<E> implements List<E> {

    /**
     * Note that we have to initialize the array an array of Objects, which
     * is the base class for all Java objects - and because of this we can store
     * our objects of any type. 
     * 
     * Still, we are casting initially a parent class (Object) into it's child class (E)
     * which is not valid and interpreter will complain (for this reason, we're suppressing
     * the warning).
     * 
     * The reason why this is justified, is because our implementation guarantees that
     * we will not allow the ArrayList to return the element of the array before assigning
     * the correct E object to a position accessed in the array.
     */
    @SuppressWarnings("unchecked")
    private E[] array = (E[])new Object[1];
    private int size = 0;

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    /**
     * For explanation see solutions of Exercise 01.
     */
    @Override
    public void add(E newElem) {
        if (size == array.length)
            array = Arrays.copyOf(array, array.length * 2);
        array[size++] = newElem;
    }

    /**
     * For explanation see solutions of Exercise 01.
     */
    @Override
    public void remove(int index) {
        checkIndex(index);
        System.arraycopy(array, index + 1, array, index, size - 1 - index);
        array[--size] = null;
    }

    @Override
    public E get(int index) {
        checkIndex(index);
        return array[index];
    }

    @Override
    public void set(int index, E elem) {
        checkIndex(index);
        array[index] = elem;
    }

    /**
     * Implementation of the Iterable interface and anonymous implementation of
     * the Iterator<E> interface.
     */
    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            private int pos = 0;
            private boolean canRemove = false;
            public boolean hasNext() {
                return pos < size();
            }
            public E next() {
                if (pos >= size)
                    throw new NoSuchElementException();
                canRemove = true;
                return array[pos++];
            }

            /**
             * Can not remove an item unless the current node is clearly defined.
             * Illegal states are:
             *   - before the first step of the iterator
             *   - after any removal (has to perform a step before the next removal)
             *   - iterator is out of elements
             */
            public void remove() {
                if (!canRemove)
                    throw new IllegalStateException();
                ArrayList.this.remove(--pos);
                canRemove = false;
            }
        };
    }

    private void checkIndex(int index) {
        if (! (0 <= index && index < size))
            throw new IndexOutOfBoundsException("invalid index: "+ index +" (must be in [0;"+ size +"[)");
    }
}
