package ch.epfl.collections;

import java.util.Iterator;
import java.util.NoSuchElementException;

/**
 * Double-linked circular implementation of a generic List.
 *
 * Important parts:
 *  
 * @see LinkedList.Node implementation at the end
 * @see LinkedList#iterator()
 *
 */
public class LinkedList<E> implements List<E> {
    private Node<E> head = null;
    private int size = 0;

    @Override
    public boolean isEmpty() {
        return size == 0;
    }

    @Override
    public int size() {
        return size;
    }

    /**
     * Inserts a new element and links it with the existing nodes.
     * For details see Exercise 01 solutions.
     */
    @Override
    public void add(E newElem) {
        final Node<E> newNode = new Node<E>(newElem);
        if (size == 0) {
            assert head == null;
            head = newNode;
            newNode.next = newNode.prev = newNode;
        } else {
            assert head != null;
            Node<E> tail = head.prev;
            tail.next = newNode;
            newNode.prev = tail;
            newNode.next = head;
            head.prev = newNode;
        }
        size += 1;
    }

    @Override
    public void remove(int index) {
        removeNode(nodeAtIndex(index));
    }

    @Override
    public E get(int index) {
        return nodeAtIndex(index).value;
    }

    @Override
    public void set(int index, E elem) {
        nodeAtIndex(index).value = elem;
    }

    /**
     * Implementation of the Iterable interface and anonymous implementation of
     * the Iterator<E> interface.
     */
    @Override
    public Iterator<E> iterator() {
        return new Iterator<E>() {
            private Node<E> nextNode = head;
            private int nbLeft = size;
            private boolean canRemove = false;

            public boolean hasNext() {
                return nbLeft > 0;
            }

            public E next() {
                if (nbLeft <= 0)
                    throw new NoSuchElementException();
                E elem = nextNode.value;
                nextNode = nextNode.next;
                --nbLeft;
                canRemove = true;
                return elem;
            }

            /**
             * Can not remove an item unless the current node is clearly defined.
             * Illegal states are:
             *   - before the first step of the iterator
             *   - after any removal (has to perform a step before the next removal)
             *   - iterator is out of elements
             */
            public void remove() {
                if (!canRemove) 
                    throw new IllegalStateException();
                removeNode(nextNode.prev);
                canRemove = false;
            }
        };
    }

    private Node<E> nodeAtIndex(int index) {
        if (index >= size)
            throw new IndexOutOfBoundsException("invalid index: "+ index +" (must be in [0;"+ (size() - 1) + "])");

        Node<E> n = head;
        for (int i = 0; i < index; ++i)
            n = n.next;
        return n;
    }

    /**
     * Internal method for removing the node.
     * 
     * Updates the .next and .prev references of the remaining nodes in the chain.
     * 
     * @param node Node to be removed.
     */
    private void removeNode(final Node<E> node) {
        node.prev.next = node.next;
        node.next.prev = node.prev;
        --size;
        if (node == head) // in case we're removing the head node, update reference to the new head
            head = (size == 0 ? null : head.next);
    }

    /**
     * Generic node class, used only internally in the LinkedList class.
     *
     */
    private static class Node<E> {
        E value;
        Node<E> prev = null, next = null;

        public Node(E value) {
            this.value = value;
        }
    }
}
